#include "file/file_intel_hex.h"
#include "common/crc.h"

#include <iostream>
#include <iomanip>

#include <fstream>
#include <sstream>

#include <locale>

#include <algorithm>
#include <vector>

FileIntelHex::Error_state FileIntelHex::m_error_state = FileIntelHex::Error_state::none;
std::string FileIntelHex::m_error_text = "";
std::string FileIntelHex::m_error_detail = "";

FileIntelHex::FileIntelHex() :
    m_max_data_size( 16 ),
    m_fill_byte( 0xFF ),
    m_block_size( 0x10000 )
{
    // Der Standardmäßig zu verarbeitende Datensatztyp
    setRecordType( Record_type::data );

    // Für die CRC-Berechnung Initalisieren.
    InitCRC32Table();

    // Initialisieren
    resetData();

    // Fehlerobjekt zuruecksetzen
    resetError();
}

FileIntelHex::~FileIntelHex()
{
}

void FileIntelHex::resetData()
{
    // Angabe der Datei
    setFilePath( "" );

    // Register und Puffer
    m_start_segment_address.ipRegister = 0;
    m_start_segment_address.csRegister = 0;
    m_start_segment_address.exists = false;

    m_start_linear_address.eipRegister = 0;
    m_start_linear_address.exists = false;

    m_base_address = 0;
    m_address_mode = Address_mode::unknown;
    m_byte_marker.clear();
    m_buffer.clear();
}

void FileIntelHex::resetError()
{
    // Member initialisieren
    m_error_state = Error_state::none;
    m_error_text.clear();
    m_error_detail.clear();
}

bool FileIntelHex::load( const std::string &filePath )
{
    // Lesen der Hexdatei in die internen Speicher.
    std::ifstream fileIn( filePath );
    std::string lineBuffer;

    bool initAddresses = false;
    unsigned int numLine = 0;
    unsigned int extendedAddress = 0;
    unsigned int maxAddress = 0;
    unsigned short recordAddress;
    unsigned char byteCount;
    unsigned char recType;

    // Datenumgebung zuruecksetzen.
    resetData();

    // Fehlerobjekt zuruecksetzen
    resetError();

    if ( fileIn.is_open() == false )
    {
        // Oeffnen der Datei gescheitert
        setErrorState( Error_state::file_open );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
    }
    else
    {
        // Zeilenweise lesen, bis eof oder Fehler.
        while( fileIn.eof() == false && (errorState() == Error_state::none) )
        {
            lineBuffer = nextDataLine( fileIn );

            // Zeile lesen.
            if ( lineBuffer.size() >= INTELHEX_STRING_LENGTH_FRAME )
            {
                // Anzahl der Datenzeilen.
                numLine++;

                recType = GetType( lineBuffer.data() );
                // Je nach Datensatztyp verarbeiten.
                switch ( static_cast<Record_type>( recType ) )
                {
                case Record_type::data:                // Data record
                case Record_type::display_font:        // Datei der Anzeigedaten
                case Record_type::display_grafic:
                case Record_type::device_header:       // Datei der Geraetedaten
                case Record_type::text:           // Datei der Textdaten
                case Record_type::touch_key:       // Datei der Touchkeydaten
                case Record_type::display_color:   // Datei der Displayfarben
                case Record_type::wlan_config:     // Datei der WlanConfig
                case Record_type::hardware_config: // Datei der HardwareConfig
                    {
                        // Nur diesen Datensatztypen verarbeiten.
                        if ( recordType() != static_cast<Record_type>( recType ) )
                            continue;

                        recordAddress = GetAddress( lineBuffer.data() );
                        byteCount = GetCount( lineBuffer.data() );

                        // Erkennung ob bereits die Basisaddresse initalisiert ist.
                        if ( !initAddresses )
                        {
                            // Basisadresse für die Daten halten.
                            m_base_address = extendedAddress + recordAddress;
                            initAddresses = true;
                        }

                        // Wird ggf. vor der bisherigen Adresse eingefügt.
                        if ( m_base_address > (extendedAddress + recordAddress) )
                        {
                            // Diese Daten müssen vor den bisherigen eingefügt werden.

                            // Diese Addresse wurde bereits beschrieben.
                            setErrorState( Error_state::file_format );
                            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                            /*
                            // Die bisherigen Daten müssen um die byteCount nach hinten verschoben werden.
                            bufferFillBytes_.resize( tempBuffer.size() );
                            std::fill_n( tempFillBytes.begin(), tempFillBytes.size(), false );
                            bufferFillBytes_.insert( bufferFillBytes_.begin(), tempFillBytes.begin(), tempFillBytes.end() );

                            std::fill_n( tempBuffer.begin(), tempBuffer.size(), fillByte() );
                            buffer_.insert( buffer_.begin(), tempBuffer.begin(), tempBuffer.end() );

                            // Basisadresse neu setzen.
                            baseAddress_ = extendedAddress + recordAddress;
                            */
                        }
                        // Eventuell nächsten Speicherblock reservieren.
                        if ( ((extendedAddress + recordAddress + byteCount) - m_base_address) > m_buffer.size() )
                        {
                            m_byte_marker.resize(m_byte_marker.size() + blockSize());
                            m_buffer.resize( m_buffer.size() + blockSize(), fillByte() );
                        }

                        GetData( lineBuffer.data(), &m_buffer[(extendedAddress + recordAddress) - m_base_address] );

                        for (unsigned int address = (extendedAddress + recordAddress) - m_base_address; address < ((extendedAddress + recordAddress) - m_base_address) + byteCount; ++address)
                        {
                            if (m_byte_marker.test(address))
                            {
                                // Diese Addresse wurde bereits beschrieben.
                                setErrorState( Error_state::file_format );
                                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                                break; // Abbruch der for-loop
                            }
                            else
                            {
                                m_byte_marker.set(address, true);
                            }
                        }

                        if ( errorState() == Error_state::none )
                        {
                            // Maximal bearbeitete Adresse halten.
                            if ( (extendedAddress + recordAddress + byteCount) > maxAddress )
                                maxAddress = extendedAddress + recordAddress + byteCount;
                        }
                        break;}

                case Record_type::end_of_file: // End-of-file record
                    break;

                case Record_type::extended_segment_address:{ // Extended Segment Address Record
                    if ( addressMode() == Address_mode::unknown )
                    {
                        setAddressMode( Address_mode::segment );
                    }
                    else if ( addressMode() != Address_mode::segment )
                    {
                        // Es geht nur entweder Segmentiert oder Linear.
                        setErrorState( Error_state::file_format );
                        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                        break;
                    }
                    // Dieses ist nun die aktuelle Basisadresse für weitere Daten (TypeDAT)
                    extendedAddress = GetData16( lineBuffer.data(), 0 ) << 4;
                    break;}

                case Record_type::start_segment_address: // Start Segment Address Record
                    if ( addressMode() == Address_mode::unknown )
                    {
                        setAddressMode( Address_mode::segment );
                    }
                    else if ( addressMode() != Address_mode::segment )
                    {
                        // Es geht nur entweder Segmentiert oder Linear.
                        setErrorState( Error_state::file_format );
                        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                        break;
                    }
                    if ( m_start_linear_address.exists )
                    {
                        // Start Linear Address ist bereits definiert, kann nur einmal geschehen.
                        setErrorState( Error_state::file_format );
                        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                        break;
                    }
                    m_start_segment_address.csRegister = GetData16( lineBuffer.data(), 0 );
                    m_start_segment_address.ipRegister = GetData16( lineBuffer.data(), 2 );
                    m_start_segment_address.exists = true;
                    break;

                case Record_type::extended_linear_address: { // Extended Linear Address Record
                    if ( addressMode() == Address_mode::unknown )
                    {
                        setAddressMode( Address_mode::linear );
                    }
                    else if ( addressMode() != Address_mode::linear )
                    {
                        // Es geht nur entweder Segmentiert oder Linear.
                        setErrorState( Error_state::file_format );
                        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                        break;
                    }
                    // Dieses ist nun die aktuelle Basisadresse für weitere Daten (TypeDAT)
                    extendedAddress = GetData16( lineBuffer.data(), 0 ) << 16;
                    break;}

                case Record_type::start_linear_address: // Start Linear Address Record
                    if ( addressMode() == Address_mode::unknown )
                    {
                        setAddressMode( Address_mode::linear );
                    }
                    else if ( addressMode() != Address_mode::linear )
                    {
                        // Es geht nur entweder Segmentiert oder Linear.
                        setErrorState( Error_state::file_format );
                        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                        break;
                    }

                    if ( m_start_segment_address.exists )
                    {
                        // Start Segment Address ist bereits definiert, kann nur einmal geschehen.
                        setErrorState( Error_state::file_format );
                        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                    }
                    m_start_linear_address.eipRegister = GetData16( lineBuffer.data(), 0 ) << 16;
                    m_start_linear_address.eipRegister += GetData16( lineBuffer.data(), 2 );
                    m_start_linear_address.exists = true;
                    break;
                }
            }
        }

        if ( errorState() == Error_state::none )
        {
            // Datenpuffer an die letzte Adresse anpassen.
            m_byte_marker.resize(maxAddress - m_base_address);
            m_buffer.resize( maxAddress - m_base_address );

            // Geladende Dateiangabe halten.
            setFilePath( filePath );
        }
        else
        {
            // Datenumgebung zuruecksetzen.
            resetData();
        }
    }

    return (errorState() == Error_state::none);
}

bool FileIntelHex::save( const std::string &filePath, bool appendToFile /*= false*/, bool appendEndLine /*= true */ )
{
    // Schreiben der internen Speicher in die Hex-Datei.
    std::ofstream fileOut( filePath, appendToFile ? std::ios::app : std::ios::out );
    std::string lineBuffer;

    unsigned int numLine = 0;
    unsigned int currentAddress;
    unsigned int extendedAddress = ~0;
    unsigned short recordAddress;

    std::vector<unsigned char>::iterator iterContent;
    std::vector<unsigned char> recordData;
    std::vector<unsigned char> addressData;

    // Fehlerobjekt zuruecksetzen
    resetError();

    if ( fileOut.is_open() == false )
    {
        // Oeffnen der Datei gescheitert
        setErrorState( Error_state::file_open );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
    }
    else
    {
        // Wenn keine FillBytes-Information vorliegt, dann eine Standard erzeugen und alle Bytes schreiben.
        if ( m_byte_marker.size() < m_buffer.size() )
            m_byte_marker.resize( m_buffer.size(), true );

        // Iterator auf den Datenbegin.
        iterContent = m_buffer.begin();
        // Aktuelle Adresse zuweisen.
        currentAddress = m_base_address;

        // Schreiben der Anwendungsfirmware
        for ( ; ;++iterContent, ++currentAddress )
        {
            // Datensatz bei folgenden Bedingungen schreiben.
            // - Puffer der Daten hat sein Ende erreicht, Angesammelte Daten schreiben.
            if ( iterContent == m_buffer.end() )
            {
                if ( recordData.size() )
                {
                    // Angefangenen Datenrecord schreiben
                    lineBuffer = typedDataLine( static_cast<unsigned short>( (currentAddress & 0xFFFF) - recordData.size() ), 
                        static_cast<unsigned char>( recordType() ), recordData.data(), static_cast<unsigned char>( recordData.size() ) );
                    if ( lineBuffer.size() == 0 )
                    {
                        // Fehler zuweisen
                        setErrorState( Error_state::record_make );
                        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                        break; // Abbruch der for-loop.
                    }

                    fileOut << lineBuffer << std::endl;
                    numLine++;
                }

                // Abbruchbedingung, fertig.
                break;
            }

            // Datensatz bei folgenden Bedingungen schreiben.
            // - Aktuelles Datenbyte ist nicht als gelesen markiert und soll somit auch nicht geschrieben werden, Angesammelte Daten schreiben.
            if ( m_byte_marker.test(currentAddress - m_base_address) == false )
            {
                if ( recordData.size() )
                {
                    // Angefangenen Datenrecord schreiben
                    lineBuffer = typedDataLine( static_cast<unsigned short>( (currentAddress & 0xFFFF) - recordData.size() ), 
                        static_cast<unsigned char>( recordType() ), recordData.data(), static_cast<unsigned char>( recordData.size() ) );
                    if ( lineBuffer.size() == 0 )
                    {
                        // Fehler zuweisen
                        setErrorState( Error_state::record_make );
                        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                        break; // Abbruch der for-loop.
                    }

                    fileOut << lineBuffer << std::endl;
                    numLine++;

                    // Datensatzpuffer leeren
                    recordData.clear();
                }

                // Weiter mit naechstem Byte.
                continue;
            }

            // Datensatz bei folgenden Bedingungen schreiben.
            // - Wechsel der Erweiterten Addresse für aktuelles Datenbyte, Angesammelte Daten schreiben.
            if (  recordData.size() && (currentAddress & 0xFFFF0000) != extendedAddress )
            {
                // Angefangenen Datenrecord schreiben
                lineBuffer = typedDataLine( static_cast<unsigned short>( (currentAddress & 0xFFFF) - recordData.size() ), 
                    static_cast<unsigned char>( recordType() ), recordData.data(), static_cast<unsigned char>( recordData.size() ) );
                if ( lineBuffer.size() == 0 )
                {
                    // Fehler zuweisen
                    setErrorState( Error_state::record_make );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                    break; // Abbruch der for-loop.
                }

                fileOut << lineBuffer << std::endl;
                numLine++;

                // Datensatzpuffer leeren
                recordData.clear();
            }

            // Wechsel der Erweiterten Addresse?
            if ( (currentAddress & 0xFFFF0000) != extendedAddress )
            {
                // Nun aktuelle erweiterte Adresse zuweisen.
                extendedAddress = currentAddress & 0xFFFF0000;

                // Erweiterte Adresse je nach Modus schreiben.
                switch ( addressMode() )
                {
                case Address_mode::unknown:
                    break;

                case Address_mode::segment:
                    {
                        recordAddress = (extendedAddress >> 4) & 0xFFFF;
                        addressData.clear();
                        addressData.push_back( (recordAddress >> 8) & 0xFF );
                        addressData.push_back( recordAddress & 0xFF );

                        // Extended Segment Address Record
                        lineBuffer = typedDataLine( 0, static_cast<unsigned char>( Record_type::extended_segment_address ), addressData.data(), static_cast<unsigned char>( addressData.size() ) );
                        if ( lineBuffer.size() == 0 )
                        {  
                            // Fehler zuweisen
                            setErrorState( Error_state::record_make );
                            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                        }
                        else
                        {
                            fileOut << lineBuffer << std::endl;
                            numLine++;
                        }
                        break;
                    }

                case Address_mode::linear:
                    {
                        recordAddress = (extendedAddress >> 16) & 0xFFFF;
                        addressData.clear();
                        addressData.push_back( (recordAddress >> 8) & 0xFF );
                        addressData.push_back( recordAddress & 0xFF );

                        // Extended Segment Address Record
                        lineBuffer = typedDataLine( 0, static_cast<unsigned char>( Record_type::extended_linear_address ), addressData.data(), static_cast<unsigned char>( addressData.size() ) );
                        if ( lineBuffer.size() == 0 )
                        {  
                            // Fehler zuweisen
                            setErrorState( Error_state::record_make );
                            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                        }
                        else
                        {
                            fileOut << lineBuffer << std::endl;
                            numLine++;
                        }
                        break;
                    }

                }

                // Fehler in switch zugewiesen?
                if ( errorState() != Error_state::none )
                    break;  // Abbruch der for-loop.
            }

            // Aktuelles Datenbyte anfügen.
            recordData.push_back( (*iterContent) );

            // Datensatz bei folgenden Bedingungen schreiben.
            // - Maximale Anzahl Datenbyte pro Datensatz erreicht.
            if ( recordData.size() == maxDataSize() )
            {
                // Angefangenen Datenrecord schreiben
                lineBuffer = typedDataLine( static_cast<unsigned short>( (currentAddress & 0xFFFF) - recordData.size() ) + 1, 
                    static_cast<unsigned char>( recordType() ), recordData.data(), static_cast<unsigned char>( recordData.size() ) );
                if ( lineBuffer.size() == 0 )
                {
                    // Fehler zuweisen
                    setErrorState( Error_state::record_make );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                    break; // Abbruch der for-loop.
                }

                fileOut << lineBuffer << std::endl;
                numLine++;

                // Datensatzpuffer leeren
                recordData.clear();
            }
        }

        // Kein Fehler aufgetreten.
        if ( errorState() == Error_state::none )
        {
            if ( m_start_segment_address.exists )
            {
                // Start Segment Address Record schreiben.

                addressData.clear();
                addressData.push_back( (m_start_segment_address.csRegister >> 8) & 0xFF );
                addressData.push_back( m_start_segment_address.csRegister & 0xFF );
                addressData.push_back( (m_start_segment_address.ipRegister >> 8) & 0xFF );
                addressData.push_back( m_start_segment_address.ipRegister & 0xFF );
                // Start Segment Address Record
                lineBuffer = typedDataLine( 0, static_cast<unsigned char>( Record_type::start_segment_address ), addressData.data(), static_cast<unsigned char>( addressData.size() ) );
                if ( lineBuffer.size() == 0 )
                {  
                    // Fehler zuweisen
                    setErrorState( Error_state::record_make );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                }
                else
                {
                    fileOut << lineBuffer << std::endl;
                    numLine++;
                }
            }
            else if ( m_start_linear_address.exists )
            {
                // Start Linear Address Record schreiben.

                addressData.clear();
                addressData.push_back( (m_start_linear_address.eipRegister >> 24) & 0xFF );
                addressData.push_back( (m_start_linear_address.eipRegister >> 16) & 0xFF );
                addressData.push_back( (m_start_linear_address.eipRegister >> 8) & 0xFF );
                addressData.push_back( m_start_linear_address.eipRegister & 0xFF );
                // Start Linear Address Record
                lineBuffer = typedDataLine( 0, static_cast<unsigned char>( Record_type::start_linear_address ), addressData.data(), static_cast<unsigned char>( addressData.size() ) );
                if ( lineBuffer.size() == 0 )
                {  
                    // Fehler zuweisen
                    setErrorState( Error_state::record_make );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                }
                else
                {
                    fileOut << lineBuffer << std::endl;
                    numLine++;
                }
            }

            if ( appendEndLine )
            {
                lineBuffer = endLine();
                if ( lineBuffer.size() == 0 )
                {
                    // Fehler zuweisen
                    setErrorState( Error_state::record_end_make );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                }
                else
                {
                    fileOut << lineBuffer << std::endl;
                    numLine++;
                }
            }

            // Geschriebene Dateiangabe halten.
            if ( errorState() == Error_state::none )
                setFilePath( filePath );
        }
    }

    return (errorState() == Error_state::none);
}

unsigned int FileIntelHex::checksum32bit()
{
    unsigned int crc = -1;

    // CRC berechnen.
    if ( m_buffer.size() )
    {
        crc = GetCRC32( m_buffer.data(), static_cast<unsigned int>( m_buffer.size() ) );
    }

    return crc;
}

unsigned int FileIntelHex::crc32( const unsigned char *data, unsigned int size, unsigned int crc, bool closure )
{
    return GetCRC32( data, size, crc, closure );
}

unsigned short FileIntelHex::crc16( const unsigned char *data, unsigned int size, unsigned short crc )
{
    return GetCRC16( data, size, crc );
}

// ########################################################
// HELPER

unsigned short FileIntelHex::HexToUINT16( const char *data )
{
    int i;
    unsigned short val = 0;

    for ( i = 0; i < 4; ++i )
    {
        val = (val << 4);

        if ( data[i] >= '0' && data[i] <= '9' )
            val += data[i] - '0';
        else if ( data[i] >= 'a' && data[i] <='f' )
            val += data[i] - 'a' + 10;
        else
            val += data[i] - 'A' + 10;
    }

    return val;
}

void FileIntelHex::UINT16ToHex( unsigned short val, char *data )
{
    int i;

    for ( i = 3; i >= 0; --i )
    {
        data[i] = val & 0xF;

        if ( data[i] > 9 )
            data[i] += 'A' - 10; 
        else
            data[i] += '0';

        val >>= 4;
    }	
}

unsigned char FileIntelHex::HexToUINT8( const char *data )
{
    int i;
    unsigned char val = 0;

    for ( i = 0; i < 2; ++i )
    {
        val = (val << 4);

        if ( data[i] >= '0' && data[i] <= '9' )
            val += data[i] - '0';
        else if ( data[i] >= 'a' && data[i] <='f' )
            val += data[i] - 'a' + 10;
        else
            val += data[i] - 'A' + 10;
    }

    return val;
}

void FileIntelHex::UINT8ToHex( unsigned char val, char *data )
{
    data[0] = (val >> 4) & 0xF;

    if ( data[0] > 9 )
        data[0] += 'A' - 10; 
    else
        data[0] += '0';

    data[1] = val & 0xF;

    if ( data[1] > 9 )
        data[1] += 'A' - 10;
    else
        data[1] += '0';
}

bool FileIntelHex::appendCrc( std::string &data )
{
    int length = 0;
    int lastByte;
    int check = 0;
    std::vector<std::string::value_type> crc( 3, '\0' );

    if ( data[INTELHEX_OFFSET_START] != INTELHEX_RECORD_MARK )
        return false;

    length++;
    // Laenge der Datenzeile ermitteln
    while( std::isxdigit( data[length], m_locale ) )
    {
        length++;

        if ( (length % 2) == 0 )
        {
            lastByte = HexToUINT8( &data[length - 1] );
            check += lastByte;
        }
    }

    if ( length < 9 ) // Mindestens 9 Zeichen: ":00000001"
        return false;
    else if ( length != (INTELHEX_STRING_LENGTH_HEAD + GetCount( &data[0] ) * 2 ) ) // Aktuell eingetragene Laaenge.
        return false;

    // Pruefsumme eintragen
    data.resize( length + 2 );
    UINT8ToHex( static_cast<unsigned char>(~(*reinterpret_cast<char *>( &check ) - 1) & 0xFF), &data[length] );

    return true;
}

bool FileIntelHex::checkCrc( const char *data )
{
    int length = 0;
    int lastByte = 0;
    int check = 0;

    if ( data[INTELHEX_OFFSET_START] != INTELHEX_RECORD_MARK )
        return false;

    length++;
    // Laenge der Datenzeile ermitteln
    while( std::isxdigit( data[length], m_locale ) )
    {
        length++;

        if ( (length % 2) == 0 )
        {
            lastByte = HexToUINT8( &data[length - 1] );
            check += lastByte;
        }
    }

    if ( length < 11 ) // Mindestens 11 Zeichen: ":00000001FF"
        return false;
    else if ( length != (INTELHEX_STRING_LENGTH_FRAME + GetCount( &data[0] ) * 2) ) // Aktuell eingetragene Laenge.
        return false;

    // Pruefsumme aus Gesamtsumme entfernen
    check -= lastByte;
    // Zweierkomplement
    check = ~(((char*)&check)[0] - 1) & 0xFF;

    return (check == lastByte);
}

std::string FileIntelHex::typedDataLine( unsigned short address, unsigned char type, unsigned char *data, unsigned char length )
{
    std::string res( INTELHEX_STRING_LENGTH_HEAD, '\0' );
    res[INTELHEX_OFFSET_START] = INTELHEX_RECORD_MARK;     // start of record
    UINT8ToHex( length, &res[INTELHEX_OFFSET_LENGTH] );    // byte count
    UINT16ToHex( address, &res[INTELHEX_OFFSET_OFFSET] );  // load address
    UINT8ToHex( type, &res[INTELHEX_OFFSET_TYPE] );        // record type

    res.resize( res.size() + length * 2 );
    // insert data
    unsigned char i;
    for ( i = 0; i < length; ++i )
        UINT8ToHex( data[i], &res[INTELHEX_OFFSET_DATA + (i * 2)] );

    // checksum suffix
    if ( appendCrc( res ) == false )
    {
        // Fehler zuweisen
        setErrorState( Error_state::crc_fail );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        res = "";
    }

    return res;
}

std::string FileIntelHex::dataLine( unsigned short address, unsigned char *data, unsigned char length )
{
    return typedDataLine( address, static_cast<unsigned char>( Record_type::data ), data, length );
}

std::string FileIntelHex::endLine()
{
    std::string res( INTELHEX_STRING_LENGTH_HEAD, '\0' );
    res[INTELHEX_OFFSET_START] = INTELHEX_RECORD_MARK; // start of record
    UINT8ToHex( 0, &res[INTELHEX_OFFSET_LENGTH] );     // byte count
    UINT16ToHex( 0, &res[INTELHEX_OFFSET_OFFSET] );    // load address
    UINT8ToHex( static_cast<unsigned char>( Record_type::end_of_file ), &res[INTELHEX_OFFSET_TYPE] ); // record type

    // checksum suffix
    if ( appendCrc( res ) == false )
    {
        // Fehler zuweisen
        setErrorState( Error_state::crc_fail );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        res = "";
    }

    return res;
}

// ########################################################
// Sucht naechsten Datensatz und fuehrt grundlegende Pruefungen mittels checkCrc durch.
std::string FileIntelHex::nextDataLine( std::istream &s )
{
    std::string res(INTELHEX_STRING_LENGTH_DATA, 0);
    char c;

    // Eine Datenzeile ermitteln.
    do
    {
        c = s.peek();

        if ( s.eof() || s.fail() )
            break; // Abbruch der while-loop

        if ( c == '\n' || c == '\a' || c == '\r' || c == ' ' || c == '\t' )
        {
            s.get(); // Zeichen zwischen den Datensaetzen uebergehen.
        }
        else if ( c == INTELHEX_RECORD_MARK )
        {
            res[0] = s.get();// Beginn des Datensatzes.
            s.get(&res[1], res.size() - 1, INTELHEX_RECORD_MARK);
            break;
        }
        else
        {
            // Unerlaubte Zeichen zwischen den Datensaetzen.
            setErrorState( Error_state::record_format );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            break; // Abbruch der while-loop
        }
    }
    while ( true );

    if ( errorState() == Error_state::none )
    {
        // Daten holen.
        std::string::size_type pos = res.find('\0', INTELHEX_STRING_LENGTH_FRAME);
        if (pos == std::string::npos)
            pos = 0;

        res.resize(pos);
        if ( res.size() > 0 && checkCrc( res.data() ) == false )
        {
            setErrorState( Error_state::record_crc );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            res = "";
        }
    }
    else
    {
        res = "";
    }

    return res;
}

unsigned char FileIntelHex::GetCount( const char *data )
{
    return HexToUINT8( &data[INTELHEX_OFFSET_LENGTH] );
}

unsigned short FileIntelHex::GetAddress( const char *data )
{
    return HexToUINT16( &data[INTELHEX_OFFSET_OFFSET] );
}

unsigned char FileIntelHex::GetType( const char *data )
{
    return HexToUINT8( &data[INTELHEX_OFFSET_TYPE] );
}

unsigned char FileIntelHex::GetData( const char *data, unsigned char *dest )
{
    // vom Datensatztyp unabhaengig
    unsigned short num, length = GetCount( data );
    if ( length > 0xFF )
        return 0;

    for ( num = 0; num < length; ++num )
        dest[num] = HexToUINT8( &data[INTELHEX_OFFSET_DATA + 2 * num] );

    return (unsigned char)num;
}

unsigned char FileIntelHex::GetData8( const char *data, unsigned char num )
{
    if ( num > GetCount( data ) )
        return 0;

    return HexToUINT8( &data[INTELHEX_OFFSET_DATA + 2 * num] );
}

unsigned short FileIntelHex::GetData16( const char *data, unsigned char num )
{
    if ( (num + 1) > GetCount( data ) )  
        return 0;

    return HexToUINT16( &data[INTELHEX_OFFSET_DATA + 2 * num] );
}

std::string FileIntelHex::errorString( Error_state state )
{
    std::string text = "Unbekannter Fehlerzustand.";

    switch ( state )
    {
    case Error_state::none:                  text = "Kein Fehler im Zustand."; break;
    case Error_state::file_open:              text = "Oeffnen der Datei gescheitert."; break;
    case Error_state::record_empty:           text = "Die uebergebene Datenzeile ist leer."; break;
    case Error_state::record_crc:             text = "Der Berechnete CRC stimmt nicht mit dem im Datensatz ueberein."; break;
    case Error_state::record_format:          text = "Der Datensatz weist fehler auf. Puffer zu kurz, lang oder fehlerhafte Laengenabgabe."; break;
    case Error_state::record_buffer_overflow:  text = "Der Datensatz geht ueber die Grenze des uebergebenen Puffers hinweg."; break;
    case Error_state::file_write:             text = "Es ist ein Dateifehler beim Schreiben der Daten aufgetreten."; break;
    case Error_state::file_read:              text = "Es ist ein Dateifehler beim Lesen der Daten aufgetreten."; break;
    case Error_state::file_empty:             text = "In der Datei liegt kein Datensatz vor."; break;
    case Error_state::crc_fail:               text = "CRC von Daten falsch."; break;
    case Error_state::format:                text = "Bereichsfehler bei eingelesenen Daten usw.."; break;
    case Error_state::no_data:                text = "Keine Daten beim laden einer Datei."; break;
    case Error_state::load_display_data:       text = "Laden der Anzeigedaten gescheitert."; break;
    case Error_state::load_device_text:        text = "Laden der Anzeigetexte gescheitert."; break;
    case Error_state::device_unknown:         text = "Unbekannter Geraetetyp der Geraetedatei."; break;
    case Error_state::program_overflow:       text = "Programmspeicher Ueberlauf."; break;
    case Error_state::record_make:            text = "Erzeugen eines IntelHex Datensatzes ist fehlgeschlagen."; break;
    case Error_state::crypt_fix_fail:          text = "Verschluesselungswert ungueltig."; break;
    case Error_state::crypt_none:             text = "Entschluesselungsversuch ohne vorherige Verschluesselung."; break;
    case Error_state::crypt_is:               text = "Verschluesselungsversuch obwohl bereits verschluesselt."; break;
    case Error_state::ini_key:                text = "Schluesselwert aus Inidatei konnte nicht geladen werden."; break;
    case Error_state::ini_key_value:           text = "Schluesselwert aus Inidatei liegt ausserhalb des gueltigen Berreiches."; break;
    case Error_state::size:                  text = "Groesse der moeglichen Datenmenge ueberschritten."; break;
    case Error_state::version:               text = "Version der Geraetedatei ist aktueller als erwartet."; break;
    case Error_state::record_end_make:         text = "Daten oder Endzeile fuer Hexdatei konnte nicht erstellt werden."; break;
    case Error_state::head_crc:               text = "Der CRC fuer den Dateikopf schlug fehl."; break;
    case Error_state::firmware_version_length: text = "Zeichenkette fuer die Firmwareversion ist zu kurz (< 8) oder zu lang (> 32)."; break;
    case Error_state::program_crc:            text = "Der CRC fuer den Programm-Flash schlug fehl."; break;
    case Error_state::function_unavailable:   text = "Funktion wird nicht unterstuetzt."; break;
    case Error_state::load_raw_firmware:       text = "Laden der Rohen Firmwaredaten gescheitert."; break;
    case Error_state::version_file:           text = "Keine gueltige Dateiversionsangabe vorhanden."; break;
    case Error_state::as_text:                text = "Fehlertext steht ueber string zur Verfuegung."; break;
    case Error_state::file_format:            text = "Das Dateiformat ist nicht wie erwartet Intel-Hex."; break;
    case Error_state::line_format:            text = "Das Format einer Datenzeile ist falsch."; break;
    case Error_state::alloc_mem:              text = "Speicherallokation konnte nicht durchgefuehrt werden."; break;
    case Error_state::no_file_match:           text = "Es wurde keine zutreffende Geraetedatei in diesem Archiv gefunden."; break;
    case Error_state::serialize_file_data:     text = "Verarbeitung der Dateidaten ist gescheitert."; break;
    default:                                    text = "Unbekannter Fehlerzustand."; break;
    }

    return boost::str( boost::format( "(%1%): %2%" ) % static_cast<int>( state ) % text );
}

